#define sink_width 48
#define sink_height 48
short sink_bits[] = {
   0xffff, 0xffff, 0xffff, 0xffff,
   0xffff, 0x9f80, 0xffff, 0xffff,
   0x9f9f, 0xffff, 0xffff, 0x8000,
   0xffff, 0x7fff, 0xbffe, 0xffff,
   0x7fff, 0xa003, 0xffff, 0x7fff,
   0xaffd, 0xffff, 0x3fff, 0xaff9,
   0xffff, 0xffff, 0xafff, 0xffff,
   0xffff, 0xaffc, 0xffff, 0x7fff,
   0xaff8, 0xffff, 0xffff, 0xaffc,
   0xffff, 0xffff, 0xafff, 0xffff,
   0xbfff, 0xaff7, 0xffff, 0x3fff,
   0xaff3, 0xffff, 0xffff, 0xaffc,
   0x003f, 0x0000, 0x2000, 0x007f,
   0x0000, 0xe000, 0xf8df, 0xffff,
   0x07ff, 0xf9cf, 0xff0f, 0xe7ff,
   0xf9cf, 0xfff7, 0xe7ff, 0xf9ff,
   0x63f7, 0xe7fb, 0xf9ff, 0x5a37,
   0xe7fb, 0xf9cf, 0x5af7, 0xe7fb,
   0xf9cf, 0x5af7, 0xe7f9, 0xf9ef,
   0xdb0f, 0xe7fa, 0xf9ff, 0xffff,
   0xe7ff, 0xf9df, 0xffff, 0xe7ff,
   0x19cf, 0xfffc, 0xe7ff, 0xd9cf,
   0xffff, 0xe7ff, 0xd9ff, 0xce47,
   0xe673, 0x19ff, 0xb5b6, 0xe7ad,
   0xd9cf, 0xb5b7, 0xe67d, 0xd9c7,
   0xb5b7, 0xe5ed, 0x19ef, 0x4db4,
   0xe673, 0xf1ff, 0xffff, 0xe3ff,
   0x03ff, 0x0380, 0xf000, 0x07ef,
   0x0100, 0xf800, 0xffc7, 0xf93f,
   0xffff, 0xffe7, 0xfd7f, 0xffe0,
   0xffff, 0x7d7f, 0xffdf, 0xffff,
   0xbd7f, 0xffb1, 0xffff, 0xbb7f,
   0xffae, 0xffef, 0xdaff, 0xffae,
   0xffc7, 0x66ff, 0xffaf, 0xffe7,
   0xbdff, 0xffaf, 0xffff, 0xc3ff,
   0xffaf, 0xffff, 0xffff, 0xffaf};
