;; Extensions

;;; Thing mouse support
(load-library "epoch-thing")

;;; Sexp mouse support
(global-set-mouse mouse-left mouse-control 'mouse-drag-sexp)
(global-set-mouse mouse-left mouse-control-up 'end-mouse-drag)

;;; Emerge epoch support
(autoload 'emerge-buffers "epoch-patch"
  "Run Emerge on two buffers."
  t)
(autoload 'emerge-buffers-with-ancestor "epoch-patch"
  "Run Emerge on two buffers, giving another buffer as the ancestor."
  t)
(autoload 'emerge-files "epoch-patch"
  "Run Emerge on two files."
  t)
(autoload 'emerge-files-with-ancestor "epoch-patch"
  "Run Emerge on two files, giving another buffer as the ancestor."
  t)

